import math
from rasterio.transform import xy
from rasterio import warp
from rasterio.crs import CRS

CRS_LATLON = CRS.from_epsg(4326)


def _utm_zone(lat, lon):
    """
    Calculates UTM zone for latitude and longitude
    :param lat:
    :param lon:
    :return: UTM zone in format 'EPSG:32XYZ'
    """
    zone = (math.floor((lon + 180)/6) % 60) + 1
    str_zone = str(zone).zfill(2)
    
    if lat > 0:
        return CRS.from_string('EPSG:326' + str_zone)
    else:
        return CRS.from_string('EPSG:325' + str_zone)


def get_utm_zone(crs, transform, shape):
    """
    Calculates the UTM zone for the image center
    :param crs: image crs
    :param transform: image transform
    :param shape: image size [height, width]
    :return: rasterio CRS of given UTM zone
    """
    # find image extents
    # todo: check for image size!
    # if it is more than 600 km longitude, or crosses the equator, or 80N/ 80S line
    # - recommend not to transform to utm at once

    # find image center
    center_xy = xy(transform, shape[0] / 2, shape[1] / 2)
    center_latlon = warp.transform(crs, CRS_LATLON, [center_xy[0]], [center_xy[1]])
    #calc zone
    return _utm_zone(center_latlon[1][0], center_latlon[0][0])



import os
import re
import glob
from warnings import warn


def parse_directory(directory, names, extensions=('tif', 'tiff', 'TIF', 'TIFF')):
    """
    Extract necessary filenames
    Args:
        directory: str
        names: tuple of str, band or file names, e.g. ['RED', '101']
        extensions: tuple of str, allowable file extensions

    Returns:
        list of matched paths
    """
    paths = glob.glob(os.path.join(directory, '*'))
    extensions = '|'.join(extensions)
    res = []
    for name in names:
        # the channel name must be either full filename (that is, ./RED.tif) or a part after '_' (./dse_channel_RED.tif)
        sep = os.sep if os.sep != '\\' else '\\\\'
        pattern = '.*(^|{}|_)({})\.({})$'.format(sep, name, extensions)
        band_path = [path for path in paths if re.match(pattern, path) is not None]

        # Normally with our datasets it will never be the case, and may indicate wrong file naming
        if len(band_path) > 1:
            warn(RuntimeWarning(
                "There are multiple files matching the channel {}. "
                "It can cause ambiguous behavior later.".format(name)))
        res += band_path

    return res




import json
import rtree
import warnings
import shapely
import shapely.geometry
from shapely.geometry import Polygon, MultiPolygon, GeometryCollection
from shapely.ops import orient

from rasterio.warp import transform_geom
from rasterio.crs import CRS
from rasterio.errors import CRSError

from ..coords import _utm_zone, CRS_LATLON


class Feature:
    """
    Proxy class for shapely geometry, include crs and properties of feature
    """

    def __init__(self, geometry, properties=None, crs=CRS_LATLON):
        self.crs = crs
        self._geometry = self._valid(
            shapely.geometry.shape(geometry))
        self.properties = properties

    def __repr__(self):
        print('CRS: {}\nProperties: {}'.format(self.crs, self.properties))
        return repr(self._geometry)

    def __getattr__(self, item):
        return getattr(self._geometry, item)
    
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    def _valid(self, shape):
        if not shape.is_valid:
            shape = shape.buffer(0)
        return shape

    def apply(self, func):
        return Feature(func(self._geometry), properties=self.properties, crs=self.crs)

    @property
    def shape(self):
        return self._geometry

    @property
    def geometry(self):
        return shapely.geometry.mapping(self._geometry)

    def as_geojson(self, hold_crs=False):
        """ Return Feature as GeoJSON formatted dict
        Args:
            hold_crs (bool): serialize with current projection, that could be not ESPG:4326 (which is standards violation)
        Returns:
            GeoJSON formatted dict
        """
        if self.crs != CRS_LATLON and not hold_crs:
            f = self.reproject(CRS_LATLON)
        else:
            f = self

        shape = f.shape
        if shape.is_empty:
            # Empty geometries are not allowed in FeatureCollections,
            # but here it may occur due to reprojection which can eliminate small geiometries
            # This case is processed separately as orient(POLYGON_EMPTY) raises an exception
            # TODO: do not return anything on empty polygon and ignore such features in FeatureCollection.geojson
            shape = Polygon()
        else:
            try:
                shape = orient(shape)
            except Exception as e:
                # Orientation is really not a crucial step, it follows the geojson standard,
                # but not oriented polygons can be read by any instrument. So, ni case of any troubles with orientation
                # we just fall back to not-oriented version of the same geometry
                warnings.warn(f'Polygon orientation failed: {str(e)}. Returning initial shape instead',
                              RuntimeWarning)
                shape = f.shape

        f = Feature(shape, properties=f.properties)
        data = {
            'type': 'Feature',
            'geometry': f.geometry,
            'properties': f.properties
        }
        return data

    @property
    def geojson(self):
        return self.as_geojson()

    def reproject(self, dst_crs):
        new_geometry = transform_geom(
            src_crs=self.crs,
            dst_crs=dst_crs,
            geom=self.geometry,
        )
        return Feature(new_geometry, properties=self.properties, crs=dst_crs)
    
    def reproject_to_utm(self):
        lon1, lat1, lon2, lat2 = self.shape.bounds
        utm_zone = _utm_zone((lat1 + lat2)/2 , (lon1 + lon2)/2)
        return self.reproject(utm_zone)


class FeatureCollection:
    """A set of Features with the same CRS"""

    def __init__(self, features, crs=CRS_LATLON):
        self.crs = crs
        self.features = self._valid(features)

        # create indexed set for faster processing
        self.index = rtree.index.Index()
        for i, f in enumerate(self.features):
            self.index.add(i, f.bounds, f.shape)

    def __getitem__(self, item):
        return self.features[item]

    def __len__(self):
        return len(self.features)

    def _valid(self, features):
        valid_features = []
        for f in features:
            if not f.geometry.get('coordinates'):  # remove possible empty shapes
                warnings.warn('Empty geometry detected. This geometry have been removed from collection.',
                              RuntimeWarning)
            else:
                valid_features.append(f)
        return valid_features

    def apply(self, func):
        """ Applies a given function to all the Features of this FeatureColletion

        Args:
            func: A function to be applied to the Features. Must take and return shapely.geometry

        Returns:
            A new FeatureCollection with modified Features
        """
        new_features = [f.apply(func) for f in self.features]
        return FeatureCollection(new_features, crs=self.crs)

    def filter(self, func):
        features = [x for x in self if func(x)]
        return FeatureCollection(features, crs=self.crs)

    def extend(self, fc):
        for i, f in enumerate(fc):
            self.index.add(i + len(self), f.bounds)
        self.features.extend(fc.features)

    def append(self, feature):
        self.index.add(len(self) + 1, feature.bounds)
        self.features.append(feature)

    def bounds_intersection(self, feature):
        idx = self.index.intersection(feature.bounds)
        features = [self.features[i] for i in idx]
        return FeatureCollection(features, self.crs)

    def intersection(self, feature):
        proposed_features = self.bounds_intersection(feature)
        features = []
        for pf in proposed_features:
            if pf.intersection(feature).area > 0:
                features.append(pf)
        return FeatureCollection(features, self.crs)

    @staticmethod
    def _read_crs(collection):

        # if there is no defined CRS in geojson file, we folloe the standard, which says that it must be lat-lon
        if 'crs' not in collection.keys():
            return CRS_LATLON

        crs_raw = collection.get('crs', CRS_LATLON)
        crs = CRS()

        try:
            if isinstance(crs_raw, str):
                crs = CRS.from_user_input(crs_raw)
            elif isinstance(crs_raw, dict):
                if 'type' in crs_raw.keys() and 'properties' in crs_raw.keys():
                    if crs_raw['type'] == 'name':
                        crs = CRS.from_user_input(crs_raw['properties']['name'])
            # Old rasterio compatibility: a separate check for validity
            if not crs.is_valid:
                message = 'CRS {} is not supported by rasterio,' \
                          'May cause an error in further reprojection or rasterization'.format(crs)
                warnings.warn(message, RuntimeWarning)
            return crs
            # Really invalid CRS will throw CRSError
        except CRSError:
            message = 'CRS was not imported correctly, assuming EPSG:4326 (lat-lon). ' \
                      'May cause an error in further reprojection or rasterization if it is not so.'
            warnings.warn(message, RuntimeWarning)
            return CRS_LATLON

    @classmethod
    def read(cls, fp):
        r"""Reading the FeatureCollection from a geojson file.
            Args:
                fp: file identifier to open and read the data
            Returns:
                new FeatureCollection with all the polygon data in the file
           """
        with open(fp, 'r', encoding='utf-8') as f:
            collection = json.load(f)

        '''
        We want the CRS to be specified in rasterio-compatible way so that we could reproject the collection
        If it is not specified, it is OK and assumed by the geojson standard that it is CRS_LATLON
        Else it must be read from the file, and if it does not meet any known scheme, we have 2 options:
        - either throw an exception (predictable, but can fail unnecessarily)
        - or ignore the CRS data with a warning (will work better normally, but can give unexpected results)
        '''
        crs = cls._read_crs(collection)
        features = []
        for i, feature in enumerate(collection['features']):
            try:
                feature_ = Feature(
                    geometry=feature['geometry'], 
                    properties=feature['properties'],
                    crs=crs,
                )
                features.append(feature_)
            except (KeyError, IndexError) as e:
                message = 'Feature #{} have been removed from collection. Error: {}'.format(i, str(e))
                warnings.warn(message, RuntimeWarning)
                
        return cls(features, crs=crs)

    def save(self, fp, indent=None, hold_crs=False):
        """ Saves feature collection as GeoJSON file
        Args:
            fp (str): filepath
            indent (int): JSON block indent
            hold_crs (bool): make GeoJSON with current projection, that could be not ESPG:4326 (which is standards violation)
        Returns:
            None
        """
        with open(fp, 'w') as f:
            json.dump(self.as_geojson(hold_crs), f, indent=indent)

    def as_geojson(self, hold_crs=False):
        """ Returns feature collection as GeoJSON string
        Args:
            hold_crs (bool): make GeoJSON with current projection, that could be not ESPG:4326 (which is standards violation)
        Returns:
            GeoJSON string
        """
        if hold_crs:
            data = {
                'type': 'FeatureCollection',
                'crs': self.crs.to_dict(),
                'features': [f.as_geojson(hold_crs=True) for f in self.features]
            }
        else:
            data = {
                'type': 'FeatureCollection',
                'crs': CRS_LATLON.to_dict(),
                'features': [f.as_geojson() for f in self.features]
            }
        return data

    @property
    def geojson(self):
        return self.as_geojson()

    def reproject(self, dst_crs):
        """
        Reprojects all the features to the new crs
        Args:
            dst_crs: rasterio.CRS or any acceptable by your rasterio version input (str, dict, epsg code), ot 'utm'

        Returns:
            new reprojected FeatureCollection
        """
        if isinstance(dst_crs, str) and dst_crs == 'utm':
            lon1, lat1, lon2, lat2 = self.index.bounds
            dst_crs = _utm_zone((lat1 + lat2)/2, (lon1 + lon2)/2)
        else:
            dst_crs = dst_crs if isinstance(dst_crs, CRS) else CRS.from_user_input(dst_crs)

        # Old rasterio compatibility: a separate check for validity
        if not dst_crs.is_valid:
            raise CRSError('Invalid CRS {} given'.format(dst_crs))

        features = [f.reproject(dst_crs) for f in self.features]
        return FeatureCollection(features, dst_crs)

    def reproject_to_utm(self):
        """
        Alias of `reproject` method with automatic Band utm zone determining
        The utm zone is determined according to the center of the bounding box of the collection.
        Does not suit to large area geometry, that would not fit into one zone (about 6 dergees in longitude)
        """
        return self.reproject(dst_crs='utm')



import numpy as np
from rasterio.features import geometry_mask
from ..raster import BandSample


def rasterize(feature_collection, transform, out_shape, name='mask'):
    """Transform vector geometries to raster form, return band sample where
       raster is np.array of bool dtype (`True` value correspond to objects area)

    Args:
        feature_collection: `FeatureCollection` object
        transform: Affine transformation object
            Transformation from pixel coordinates of `source` to the
            coordinate system of the input `shapes`. See the `transform`
            property of dataset objects.
        out_shape: tuple or list
            Shape of output numpy ndarray.
        name: output sample name, default `mask`

    Returns:
        `BandSample` object
    """
    if len(feature_collection) > 0:
        geometries = (f.geometry for f in feature_collection)
        mask = geometry_mask(geometries, out_shape=out_shape, transform=transform, invert=True).astype('uint8')
    else:
        mask = np.zeros(out_shape, dtype='uint8')

    return BandSample(name, mask, feature_collection.crs, transform)




import numpy as np


class RandomDataset:

    def __init__(self, band_collections, sample_size, input_channels, output_labels,
                 transform=None, seed=42):
        """
        Produce samples from band collections randomly (uniform dist.) with specified shape.

        Args:
            band_collections: list of 'BandCollection' objects
            sample_size: spatial resolution of sample in pixels (height, width)
            input_channels: list of channel names, e.g. ['RED', 'GRN', '100'], case sensitive
            output_labels: list of output label names, e.g. ['100', '101', 'roof'], case sensitive
            transform: function for sample transformation
            seed: `NotImplemented`
        """
        self.band_collections = band_collections
        self.sample_size = sample_size
        self.input_channels = input_channels
        self.output_labels = output_labels
        self.transform = transform
        self.seed = seed #TODO: make dataset with seed

    def __getitem__(self, i):
        bc = self.band_collections[i]

        h, w = self.sample_size

        # random sampler
        x_range = max(1, bc.width - w)
        y_range = max(1, bc.height - h)

        x = np.random.randint(x_range)
        y = np.random.randint(y_range)

        # extract samples
        sample = dict()

        sample['image'] = (bc.sample(y, x, h, w)
                           .ordered(*self.input_channels)
                           .numpy()
                           .transpose(1, 2, 0))

        sample['mask'] = (bc.sample(y, x, h, w)
                          .ordered(*self.output_labels)
                          .numpy()
                          .transpose(1, 2, 0))

        # transform samples, e.g. augmentations or standartization
        if self.transform is not None:
            sample = self.transform(**sample)

        return sample

    def __len__(self):
        return len(self.band_collections)



import random
import string


def band_shape_guard(raster):
    raster = raster.squeeze()
    if raster.ndim != 2:
        raise ValueError('Raster file has wrong shape {}. '.format(raster.shape) +
                         'Should be 2 dim np.array')
    return raster


def random_name(length=10):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(length))


import os
import shapely

from gbdxtools import CatalogImage

from ..dataset import Feature, FeatureCollection
from .split import split as split_raster


GBDX_DIR = '/tmp/gbdx/'


def get_meta(gbdx_image):
    metadata = gbdx_image.metadata['image']

    geometry_wkt = metadata['imageBoundsWGS84']
    geometry = shapely.wkt.loads(geometry_wkt)

    properties = {
        'sat_azimuth': metadata['satAzimuth'],
        'sat_elevation': metadata['satElevation'],
        'sun_azimuth': metadata['sunAzimuth'],
        'sun_elevation': metadata['sunElevation'],
        'off_nadir': metadata['offNadirAngle'],
        'datetime': metadata['acquisitionDate'],
        'gsd': metadata['groundSampleDistanceMeters'],
        'image_id': metadata['imageId'],
        'satellite': metadata['sensorPlatformName'],
    }

    crs = gbdx_image.metadata['georef']['spatialReferenceSystemCode']
    f = Feature(geometry, properties=properties, crs=crs)
    fc = FeatureCollection([f], crs=f.crs)

    return fc


class DGImage:

    def __init__(self, image_id):
        self.image_id = image_id

        self.image = None
        self.image_path = os.path.join(GBDX_DIR, '{}.tif'.format(random_word(10)))
        self.bc = None

    def __getattr__(self, item):
        return getattr(self.image, item)

    def fetch(self, *args, **kwargs):
        self.image = CatalogImage(self.image_id, *args, **kwargs)

    def load(self, *args, **kwargs):
        if self.image is not None:
            self.image.geotiff(self.image_path, *args, **kwargs)
        else:
            raise AttributeError('Fetch image before loading')

    def transform(self, dst_path, channels):
        if os.path.exists(self.image_path):

            # split raster to separate bands
            split_raster(self.image, dst_path, channels)

            # extract meta information and save
            meta_fc = get_meta(self.image)
            meta_fc.save(os.path.join(dst_path, 'meta.geojson'))
            
        else:
            raise FileExistsError('Load GBDX Image before transform')



from ..vector import FeatureCollection


def merge_dedupe(fcs):
    """Merge feature collections of and remove duplicates"""

    def other(i, fcs):
        fc = FeatureCollection([])
        for j in range(len(fcs)):
            if i != j:
                fc.extend(fcs[j])
        return fc

    def greater(area, areas):
        for a in areas:
            if area < a:
                return False
        return True

    features = []

    for i in range(len(fcs)):

        this_fc = fcs[i]
        other_fc = other(i, fcs)

        for f in this_fc:
            inters_fc = other_fc.intersection(f)

            areas = []
            for ff in inters_fc:
                intersection = f.intersection(ff).area

                if intersection:
                    areas.append(ff.area)

            if len(inters_fc) == 0:
                features.append(f)

            elif greater(f.area, areas):
                features.append(f)

    return FeatureCollection(features, crs=fcs[0].crs)


import numpy as np

EPS = 10e-12


def __channels_flatten(arr):
    if arr.ndim == 3:
        arr = arr.reshape(-1, arr.shape[-1])
    elif arr.ndim ==2:
        arr = arr.flatten()
    return arr


def IoU(gt, pr):
    """

    Args:
        gt: raster, 2D or 3D array with shape (H, W) or (H, W, C)
        pr: raster, 2D or 3D array with shape (H, W) or (H, W, C)

    Returns:
        IoU score for each channel

    """

    if gt.ndim != pr.ndim:
        raise ValueError('Targets have different shapes: {} and {}'.format(gt.ndim, pr.ndim))

    gt = __channels_flatten(gt)
    pr = __channels_flatten(pr)

    intersection = (gt * pr).sum(axis=0)
    union = (gt + pr).sum(axis=0) - intersection

    return (intersection + EPS) / (union + EPS)


def mIoU(gt, pr, class_weights=1.):
    classes_ious = IoU(gt, pr)
    return np.mean(classes_ious * class_weights)


class GeoObject:
    """Geo spatial object base interface.

    Represents a set of metadata for the georeferenced raster object
    """
    def __init__(self):
        pass

    @property
    def crs(self):
        """ Geographic coordinate reference system of the object Returns a  `rasterio.CRS
        <https://rasterio.readthedocs.io/en/latest/api/rasterio.crs.html#rasterio.crs.CRS>`_

        """
        raise NotImplementedError

    @property
    def transform(self):
        """
        Transform matrix as the `affine.Affine
        <https://github.com/sgillies/affine>`_  object.
        This transform maps pixel row/column coordinates to coordinates in the dataset’s coordinate reference system.
        """
        raise NotImplementedError

    @property
    def res(self):
        """
        Resolution (or ground sampling distance) along X and Y axes in units of the CRS.
        Tuple (x_resolution, y_resolution)
        """
        raise NotImplementedError

    @property
    def width(self):
        """
        Width of the raster data object in pixels
        """
        raise NotImplementedError

    @property
    def height(self):
        """
        Height of the raster data object in pixels
        """
        raise NotImplementedError

    @property
    def count(self):
        """
        (int) number of channels/bands.
        """
        raise NotImplementedError

    @property
    def shape(self):
        """
        Dimensions of the raster in pixels, a tuple of int: (count, height, width); or (height, width)
        """
        raise NotImplementedError

    @property
    def nodata(self):
        """ The value that should be interpreted as \'No data\'. May be None or a value within dtype range"""
        raise NotImplementedError

    @property
    def bounds(self):
        """
        Georeferenced bounds - bounding box in the CRS of the image, based on transform and shape
        """
        raise NotImplementedError

    @property
    def profile(self):
        """ A joint representation of the main properties

        Returns:
            Dict: {
            'crs': crs,
            'nodata': nodata,
            'transform': transform
            }

        """
        return {
            'crs': self.crs,
            'nodata': self.nodata,
            'transform': self.transform,
        }

    def sample(self, y, x, height, width):
        raise NotImplementedError

    def reproject(self, dst_crs):
        raise NotImplementedError

    def reproject_to_utm(self):
        raise NotImplementedError

    def resample(self, dst_res):
        raise NotImplementedError


import cv2
import numpy as np
from collections import defaultdict
from rasterio.transform import IDENTITY, xy
from shapely.geometry import shape, Polygon, MultiPolygon, GeometryCollection

from ..vector import Feature, FeatureCollection


def polygonize(sample, epsilon=0.1, properties={}):
    """ Transform the raster mask to vector polygons.
    The pixels in the raster mask are treated as belonging to the object if their value is non-zero, and zero values are background.
    All the objects are transformed to the vector form (polygons).

    The algorithm is OpenCV
    `cv2.findContours
    <https://docs.opencv.org/3.1.0/d3/dc0/group__imgproc__shape.html#ga17ed9f5d79ae97bd4c7cf18403e1689a>`_

    This method is used as it does process the hierarchy of inlayed contours correctly.
    It also makes polygon simplification, which produces more smooth and lightweight polygons, but they do not match
    the raster mask exactly, which should be taken into account.

    Args:
        sample: BandSample to be vectorized
        epsilon: the epsilon parameter for the cv2.approxPolyDP, which specifies the approximation accuracy.
        This is the maximum distance between the original curve and its approximation
        properties: (dict) Properties to be added to the resulting FeatureCollection

    Returns:
        FeatureCollection:
            Polygons in the CRS of the sample, that represent non-black objects in the image
    """
    geoms = _vectorize(sample.numpy(), epsilon=epsilon, transform=sample.transform)
    # remove all the geometries except for polygons
    polys = _extract_polygons(geoms)
    features = ([Feature(geometry, properties=properties, crs=sample.crs)
                 for geometry in polys])
    return FeatureCollection(features, crs=sample.crs)

def _extract_polygons(geometries):
    """
    Makes the consistent polygon-only geometry list of valid polygons
    It ignores
    :return: list of shapely Polygons
    """
    shapes = []
    for geom in geometries:
        if not geom['type'] in ['Polygon', 'MultiPolygon']:
            continue
        else:
            new_shape = shape(geom).buffer(0)
            if isinstance(new_shape, Polygon):
                shapes.append(new_shape)
            elif isinstance(new_shape, MultiPolygon):
                shapes += new_shape
            elif isinstance(new_shape, GeometryCollection):
                for sh in new_shape:
                    if isinstance(sh, Polygon):
                        shapes.append(sh)
                    elif isinstance(sh, MultiPolygon):
                        shapes += sh
    return shapes


def _vectorize(binary_image, epsilon=0., min_area=1., transform=IDENTITY, upscale=1):
    """
    Vectorize binary image, returns a 4-level list of floats [[[[X,Y]]]]

    Args:
        binary_image (numpy.ndarray): image in grayscale HxW or HxWx1
        epsilon (float): approximates a contour shape to another shape with less
            number of vertices depending upon the precision we specify, %
        transform (list of float):  affine transform matrix (first two lines)
            list with 6 params - [GSDx, ROTy, ROTx, GSDy, X, Y]
        upscale (float): scale image for better precision of the polygon. The polygon is correctly downscaled back

    """
    # remove possible 1-sized dimension
    binary_image = np.squeeze(binary_image)
    if binary_image.ndim != 2:
        raise ValueError('Input image must have 2 non-degenerate dimensions')

    if upscale > 1:
        h, w = binary_image.shape[:2]
        binary_image = cv2.resize(binary_image, (int(w * upscale), int(h * upscale)), cv2.INTER_NEAREST)

    # search for all contours
    contours_result = cv2.findContours(
        binary_image,
        cv2.RETR_CCOMP, cv2.CHAIN_APPROX_TC89_KCOS)
    # For compatibility with opencv3, where contours_result[0]=image, so we should throw it away
    contours, hierarchy = contours_result if len(contours_result) == 2 else contours_result[1:]

    # approximate contours with less number of points
    if epsilon > 0.:
        contours = [cv2.approxPolyDP(cnt, epsilon, True) for cnt in contours]

    if not contours:
        return []

    cnt_children = defaultdict(list)
    child_contours = set()
    assert hierarchy.shape[0] == 1
    # http://docs.opencv.org/3.1.0/d9/d8b/tutorial_py_contours_hierarchy.html
    for idx, (_, _, _, parent_idx) in enumerate(hierarchy[0]):
        if parent_idx != -1:
            child_contours.add(idx)
            cnt_children[parent_idx].append(contours[idx])

    geometries = []
    for idx, cnt in enumerate(contours):
        if idx not in child_contours and cv2.contourArea(cnt) >= min_area:
            assert cnt.shape[1] == 1
            # take a contour as a primary shape
            coords = [[tuple(xy(transform, p[1], p[0])) for p in cnt[:, 0, :].tolist()]]
            coords[0].append(coords[0][0])
            # add all children as secondary contours (holes)
            for c in cnt_children.get(idx, []):
                if cv2.contourArea(c) >= min_area:
                    c_list = c[:, 0, :].tolist()
                    c_list.append(c_list[0])
                    coords.append([tuple(xy(transform, p[1], p[0])) for p in c_list])

            # append geometry
            geometries.append({
                'type': 'Polygon',
                'coordinates': coords,
            })
    return geometries



from functools import wraps
import keras.backend as K
from keras.utils.generic_utils import get_custom_objects


# ============================ Jaccard score ============================

def _iou_score(gt, pr, class_weights=1., smooth=1e-12):
    axes = [1, 2]

    intersection = K.sum(gt * pr, axis=axes)
    union = K.sum(gt + pr, axis=axes) - intersection
    iou = (intersection + smooth) / (union + smooth)

    # mean per image
    iou = K.mean(iou, axis=0)

    # weighted mean per class
    iou = K.mean(iou * class_weights)

    return iou


def iou_score(gt, pr):
    return _iou_score(gt, pr, class_weights=1., smooth=1e-12)


def custom_iou_score(class_weights=1, smooth=1):
    def score(gt, pr):
        return _iou_score(gt, pr, class_weights=class_weights, smooth=smooth)
    return score


# Update custom objects
get_custom_objects().update({
    'iou_score': iou_score,
})


# ============================== F-score ==============================

def _f_score(gt, pr, class_weights=1, beta=1, smooth=1e-12):
    axes = [1, 2]

    tp = K.sum(gt * pr, axis=axes)
    fp = K.sum(pr, axis=axes) - tp
    fn = K.sum(gt, axis=axes) - tp

    f_score = ((1 + beta**2) * tp + smooth) \
              / ((1 + beta**2) * tp + beta**2 * fn + fp + smooth)

    # mean per image
    f_score = K.mean(f_score, axis=0)

    # weighted mean per class
    f_score = K.mean(f_score * class_weights)

    return f_score


def f1_score(gt, pr):
    return _f_score(gt, pr, class_weights=1, beta=1, smooth=1e-12)


def custom_f_score(class_weights=1, beta=1, smooth=1):
    def score(gt, pr):
        return _f_score(gt, pr, class_weights=class_weights, beta=beta, smooth=smooth)
    return score


# Update custom objects
get_custom_objects().update({
    'f1_score': f1_score,
})



import os
import numpy as np
import warnings

import rasterio
from rasterio import Affine
from rasterio.crs import CRS
from rasterio.coords import BoundingBox
from rasterio.warp import calculate_default_transform, reproject, Resampling

from ..coords import get_utm_zone
from .geoobject import GeoObject
from ._utils import band_shape_guard, random_name

TMP_DIR = '/tmp/raster'


class Band(GeoObject):
    """Filesystem object `Band` - Rasterio DatasetReader wrapper.

    The `Band` provides access to a georeferenced raster file placed in the filesystem.
    On creation the Band opens the file for reading and
    stores all the necessary metadata and allows to read the raster data on request.

    The majority of properties are inherited from
    rasterio `DatasetReader
    <https://rasterio.readthedocs.io/en/latest/api/rasterio.io.html#rasterio.io.DatasetReader>`_.

    Any file format supported by GDAL drivers can be read.

    Args:
        fp: full path to the raster file

    """
    def __init__(self, fp):
        super().__init__()
        self._band = rasterio.open(fp)
        self._tmp_file = False

    def __del__(self):
        fp = self._band.name
        self._band.close()
        if self._tmp_file:
            os.remove(fp)

    # ======================== PROPERTY BLOCK ========================
    @property
    def crs(self):
        """
            :obj:`CRS` - coordindate reference system of the band;  in the file
        """
        return self._band.crs

    @property
    def transform(self):
        """
        Transform matrix as the `affine.Affine
        <https://github.com/sgillies/affine>`_  object.
        This transform maps pixel row/column coordinates to coordinates in the dataset’s coordinate reference system.

        affine.identity is returned if if the file does not contain transform
        """
        return self._band.transform

    @property
    def nodata(self):
        """
        Band nodata value, type depends on the image dtype; None if the nodata value is not specified
        """
        return self._band.nodata

    @property
    def res(self):
        """
        Spatial resolution (x_res, y_res) of the Band in X and Y directions of the georeferenced coordinate system,
        derived from tranaform. Normally is equal to (transform.a, - transform.e)
        """
        return self._band.res

    @property
    def width(self):
        return self._band.width

    @property
    def height(self):
        return self._band.height

    @property
    def count(self):
        """
        By design of the aeronetlib, should be always 1. A Band can be created from image of any channel count,
        but only the first band can be read. If you need to work with multi-channel image, use
        :meth:`aeronet.converters.split.split` to get the one-channel images.

        Returns:
            (int) number of the bands in the image.
        """
        return self._band.count

    @property
    def shape(self):
        """
        The raster dimension as a Tuple (height, width)
        """
        return self.height, self.width

    @property
    def name(self):
        """
        Name of the file associated with the Band, without extension and the directory path
        """
        return os.path.basename(self._band.name).split('.')[0]

    @property
    def bounds(self):
        """
        Georeferenced bounds - bounding box in the CRS of the image, based on transform and shape

        Returns:
            `BoundingBox object
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.coords.html#rasterio.coords.BoundingBox>`_:
            (left, bottom, right, top)
        """
        return self._band.bounds

    @property
    def meta(self):
        """
        The basic metadata of the associated rasterio DatasetReader
        """
        return self._band.meta

    @property
    def dtype(self):
        """
        Numerical type of the data stored in raster, according to numpy.dtype
        """
        return self._band.dtypes[0]

    # ======================== METHODS BLOCK ========================

    def numpy(self):
        """
        Read all the raster data into memory as a numpy array

        Returns:
            numpy array containing the whole Band raster data
        """
        return self.sample(0, 0, self.height, self.width).numpy()

    def same(self, other):
        """Compare if samples have same resolution, crs and shape.

        This means that the samples represent the same territory (like different spectral channels of the same image)
        and can be processed together as collection.

        Args:
            other: GeoObject to compare with

        Returns:
            True if the objects match in shape, crs, transform, False otherwise
        """
        res = True
        res = res and (self.crs == other.crs)
        res = res and (self.transform == other.transform)
        res = res and (self.height == other.height)
        res = res and (self.width == other.width)
        return res

    def _same_extent(self, other: GeoObject):
        """
        Compares the spatial extent of the current and other Bands based on their CRSes and transforms.
        The extent is treated as 'same' if the boundaries differ not more than half of the biggest pixel
        Args:
            other: GeoObject to compare the extent to
        Returns:
            bool: True if the rasters are compatible, False otherwise
        """

        # explicitly calculate the other Band's dimensions and resolution in the current crs
        other_bounds = rasterio.warp.transform_bounds(other.crs, self.crs, *other.bounds)
        other_res = [abs(other_bounds[0] - other_bounds[2])/other.width, abs(other_bounds[1] - other_bounds[3])/other.height]
        max_pixel = [max(self.res[0], other_res[0]), max(self.res[1], other_res[1])]

        # check every bound to be different not more than half of the bigger pixel
        if abs(other_bounds[0] - self.bounds[0]) > 0.5*max_pixel[0] or \
           abs(other_bounds[1] - self.bounds[1]) > 0.5 * max_pixel[1] or \
           abs(other_bounds[2] - self.bounds[2]) > 0.5 * max_pixel[0] or \
           abs(other_bounds[3] - self.bounds[3]) > 0.5 * max_pixel[1]:
            return False
        else:
            return True

    def sample(self, y, x, height, width, **kwargs):
        """ Read sample of the Band to memory.

        The sample is defined by its size and position in the raster, without respect to the georeference.
        In case if the sample coordinates spread out of the image boundaries, the image is padded with nodata value.

        Args:
            x: pixel horizontal coordinate of left top corner of the sample
            y: pixel vertical coordinate of left top corner of the sample
            width: spatial dimension of sample in pixels
            height: spatial dimension of sample in pixels

        Returns:
             a new :obj:`BandSample` containing the specified spatial subset of the band
        """

        coord_x = self.transform.c + x * self.transform.a
        coord_y = self.transform.f + y * self.transform.e

        dst_crs = self.crs
        dst_name = os.path.basename(self.name)
        dst_nodata = self.nodata if self.nodata is not None else 0
        dst_transform = Affine(self.transform.a, self.transform.b, coord_x,
                               self.transform.d, self.transform.e, coord_y)

        dst_raster = self._band.read(window=((y, y + height), (x, x + width)),
                                     boundless=True, fill_value=dst_nodata)

        sample = BandSample(dst_name, dst_raster, dst_crs, dst_transform, dst_nodata)

        return sample

    def resample(self, dst_res, fp=None, interpolation='nearest'):
        """ Change spatial resolution of the band. It does not alter the existing file,
        and creates a new file either in the specified location or a temporary file

        It is based on `rasterio.warp.reproject
        <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_,
        see for more variants of interpolation.

        Args:
            dst_res (Tuple[float, float]): new resoluton, georeferenced pixel size for the new band
            fp (str): a filename for the new resampled band. If none, a temporary file is created
            interpolation: interpolation type as in rasterio,  `nearest`, `bilinear`, `cubic`, `lanzsos` or others

        Returns:
            a new resampled Band.
        """
        # get temporary filepath if such is not provided
        tmp_file = False if fp is not None else True
        if fp is None:
            fp = '{tmp}/resampled/{directory}/{name}.tif'.format(
                tmp=TMP_DIR, directory=random_name(), name=self.name)

        os.makedirs(os.path.dirname(fp), exist_ok=True)

        transform = Affine(dst_res[0], self.transform.b, self.transform.c,
                           self.transform.d, - dst_res[1], self.transform.f)
        width = round(self.width / (dst_res[0]/self.res[0]))
        height = round(self.height / (dst_res[1]/self.res[1]))

        kwargs = self.meta.copy()
        kwargs.update({
            'transform': transform,
            'width': width,
            'height': height
        })

        with rasterio.open(fp, 'w', **kwargs) as dst:
                reproject(
                    source=rasterio.band(self._band, 1),
                    destination=rasterio.band(dst, 1),
                    src_transform=self.transform,
                    src_crs=self.crs,
                    dst_transform=transform,
                    dst_crs=self.crs,
                    resampling=getattr(Resampling, interpolation))

        # new band
        band = Band(fp)
        band._tmp_file = tmp_file  # file will be automatically removed when `Band` instance will be deleted

        return band

    def reproject(self, dst_crs, fp=None, interpolation='nearest'):
        """ Change coordinate system (projection) of the band.
        It does not alter the existing file, and creates a new file either in the specified location or a temporary file.

        The band ground sampling distance is not changed, however the resolution may change due to the new coordinate system
        It is based on `rasterio.warp.reproject
        <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_,
        see for more variants of interpolation.

        Args:
            dst_crs: new CRS, may be in any form acceptable by rasterio, for example as EPSG code, string, CRS object; if dst_crs == `utm`, the appropriate UTM zone is used according to the center of the image
            fp (str): a filename for the new resampled band. If none, a temporary file is created
            interpolation: interpolation type as in rasterio,  `nearest`, `bilinear`, `cubic`, `lanzsos` or others

        Returns:
            a new reprojected Band
        """
        if dst_crs == 'utm':
            dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        else:
            dst_crs = dst_crs if isinstance(dst_crs, CRS) else CRS.from_user_input(dst_crs)

        # Old rasterio compatibility: a separate check for validity
        if not dst_crs.is_valid:
            raise rasterio.errors.CRSError('Invalid CRS {} given'.format(dst_crs))

        # get temporary filepath if such is not provided
        tmp_file = False if fp is not None else True
        if fp is None:
            fp = '{tmp}/reprojected_{crs}/{directory}/{name}.tif'.format(
                tmp=TMP_DIR, crs=dst_crs, directory=random_name(), name=self.name)
        os.makedirs(os.path.dirname(fp), exist_ok=True)

        # calculate params of new reprojected Band
        transform, width, height = calculate_default_transform(
            self.crs, dst_crs, self.width, self.height, *self.bounds)
        kwargs = self.meta.copy()
        kwargs.update({
            'crs': dst_crs,
            'transform': transform,
            'width': width,
            'height': height
        })

        # reproject
        with rasterio.open(fp, 'w', **kwargs) as dst:
                reproject(
                    source=rasterio.band(self._band, 1),
                    destination=rasterio.band(dst, 1),
                    src_transform=self.transform,
                    src_crs=self.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    resampling=getattr(Resampling, interpolation))

        # new band
        band = Band(fp)
        band._tmp_file = tmp_file # file will be automatically removed when `Band` instance will be deleted
        return band

    def reproject_to(self, other: GeoObject, fp=None, interpolation='nearest'):
        """
        Reprojects and resamples the band to match exactly the `other`.

        This function ensures that the raster size, crs and transform will be the same,
        allowing them to be merged into one BandCollection. If the intial raster
        exceeds the other in coverage, it will be cut, and if it is insufficient or displaced, it will be zero-padded.

        It aims to overpass the rounding problem which may cause an image to
        be misaligned with itself after a different series of transforms.

        If the images are far from each other, the warning will be shown,
        because the raster may be zero due to severe misalignment.

        Args:
            other(GeoObject): the Band with the parameters to fit to
            fp (str): a filename for the new resampled band. If none, a temporary file is created
            interpolation: interpolation type as in rasterio,  `nearest`, `bilinear`, `cubic`, `lanzsos` or others.
        Returns:
            a new reprojected and resampled Band
        """
        if not self._same_extent(other):
            warnings.warn('You are trying to match two bands that are not even approxiamtely aligned. '
                          'The resulting raster may be empty')

        # get temporary filepath if such is not provided
        tmp_file = False if fp is not None else True
        if fp is None:
            fp = '{tmp}/reprojected_{crs}/{directory}/{name}.tif'.format(
                tmp=TMP_DIR, crs=other.crs, directory=random_name(), name=self.name)
        os.makedirs(os.path.dirname(fp), exist_ok=True)

        kwargs = self.meta.copy()
        kwargs.update({
            'crs': other.crs,
            'transform': other.transform,
            'width': other.width,
            'height': other.height
        })

        # reproject - as in rio.warp --like
        with rasterio.open(fp, 'w', **kwargs) as dst:
                reproject(
                    source=rasterio.band(self._band, 1),
                    destination=rasterio.band(dst, 1),
                    src_transform=self.transform,
                    src_crs=self.crs,
                    dst_transform=other.transform,
                    dst_crs=other.crs,
                    resampling=getattr(Resampling, interpolation))

        # new band
        band = Band(fp)
        band._tmp_file = tmp_file # file will be automatically removed when `Band` instance will be deleted
        return band

    def reproject_to_utm(self, fp=None, interpolation='nearest'):
        """
        Alias of :obj:`Band.reproject` method with automatic Band utm zone determining
        """
        dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        return self.reproject(dst_crs, fp=fp, interpolation=interpolation)

    def generate_samples(self, width, height):
        """
        A generator for sequential sampling of the whole band, used for the windowed reading of the raster.
        It allows to handle and process large files without reading them at once in the memory.

        Args:
            width (int): dimension of sample in pixels and step along `X` axis
            height (int): dimension of sample in pixels and step along `Y` axis

        Yields:
            BandSample: sequential samples of the specified dimensions
        """
        for x in range(0, self.width, width):
            for y in range(0, self.height, height):
                yield self.sample(y, x, height, width)


class BandSample(GeoObject):
    """ A wrapper over numpy array representing an in-memory georeferenced raster image.

    It implements all the interfaces of the GeoObject, and stores the raster data in memory

    Args:
        name (str): a name of the sample, which is used as a defaule name for saving to file
        raster (np.array): the raster data
        crs: geographical coordinate reference system, as :obj:`CRS` or string representation
        transform (Affine): affine transform for the
        nodata: the pixels with this value in raster should be ignored
    """

    def __init__(self, name, raster, crs, transform, nodata=0):
        """

        """

        super().__init__()

        self._name = name
        self._raster = band_shape_guard(raster)
        self._nodata = nodata
        self._transform = Affine(*transform) if not isinstance(transform, Affine) else transform
        self._crs = CRS.from_user_input(crs) if not isinstance(crs, CRS) else crs

        # Old rasterio compatibility: a separate check for validity
        if not self._crs.is_valid:
            raise rasterio.errors.CRSError('Invalid CRS {} given'.format(crs))

    def __eq__(self, other):
        res = np.allclose(self.numpy(), other.numpy())
        res = res and (self.crs == other.crs)
        res = res and np.allclose(np.array(self.transform), np.array(other.transform))
        return res

    def __repr__(self):
        return '<BandSample: name={}, shape={}, dtype={}>'.format(self.name,
                                                                  self.shape,
                                                                  self.dtype)

    # ======================== PROPERTY BLOCK ========================
    @property
    def width(self):
        return self._raster.shape[1]

    @property
    def height(self):
        return self._raster.shape[0]

    @property
    def count(self):
        return 1

    @property
    def shape(self):
        """
        The raster dimension as a Tuple (height, width)
        """
        return self.height, self.width

    @property
    def dtype(self):
        """
        Data type of the associated numpy array
        """
        return self._raster.dtype

    @property
    def res(self):
        return abs(self.transform.a), abs(self.transform.e)

    @property
    def transform(self):
        return self._transform

    @property
    def crs(self):
        return self._crs

    @property
    def nodata(self):
        return self._nodata

    @property
    def bounds(self):
        """
        Georeferenced bounds - bounding box in the CRS of the image, based on transform and shape

        Returns:
            `BoundingBox object
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.coords.html#rasterio.coords.BoundingBox>`_:
            (left, bottom, right, top)
        """
        left = self.transform.c
        top = self.transform.f
        right = left + self.transform.a * self.width
        bottom = top + self.transform.e * self.height
        return BoundingBox(left, bottom, right, top)

    @property
    def name(self):
        """
        name of the sample, is used as a base filename when saving to file
        """
        return self._name

    # ======================== METHODS BLOCK ========================

    @classmethod
    def from_file(cls, fp):
        """
        Reads the raster data directly from the file.
        File must have only one channel. If you need to read multi-channel file,
        use :meth:`aeronet.converters.split.split` first

        Args:
            fp: full path to the file

        Returns:
            a new `BandSample` object
        """
        band = Band(fp)
        return band.sample(0, 0, band.width, band.height)

    def same(self, other):
        """Compare if samples have same resolution, crs and shape.

        This means that the samples represent the same territory (like different spectral channels of the same image)
        and can be processed together as collection.

        Args:
            other: GeoObject to compare with

        Returns:
            True if the objects match in shape, crs, transform, False otherwise
        """

        res = True
        res = res and (self.crs == other.crs)
        res = res and (self.transform == self.transform)
        res = res and (self.height == self.height)
        res = res and (self.width == self.width)
        return res

    def save(self, directory, ext='.tif', **kwargs):
        """
        Saves the raster data to a new geotiff file; the filename is derived from this `BandSample` name.
        If file exists, it will be overwritten.

        Args:
            directory: folder to save the file
            ext: file extension; as now only GTiff driver is used, it should match `tif`, `tiff`, `TIF` or `TIFF`.
            kwargs: other keywords arguments to be passed to `rasterio.open <https://rasterio.readthedocs.io/en/latest/api/rasterio.html#rasterio.open>`_ .
        """
        fp = os.path.join(directory, self._name + ext)
        with rasterio.open(fp, mode='w', driver='GTiff', width=self.width,
                           height=self.height, count=1, crs=self.crs.to_wkt(),
                           transform=self.transform, nodata=self.nodata,
                           dtype=self.dtype, **kwargs) as dst:
            dst.write(self._raster.squeeze(), 1)

    def sample(self, y, x, height, width):
        """ Subsample of the Sample with specified dimensions and position within the raster:

        Args:
            x (int): horizontal pixel coordinate of left top corner
            y (int): vertical pixel coordinate of left top corner
            width (int): spatial x-dimension of sample in pixels
            height (int): spatial y-dimension of sample in pixels

        Return:
            a new `BandSample` object
        """

        coord_x = self.transform.c + x * self.transform.a
        coord_y = self.transform.f + y * self.transform.e

        dst_transform = Affine(self.transform.a, self.transform.b, coord_x,
                               self.transform.d, self.transform.e, coord_y)
        dst_raster = self._raster[y:y+height, x:x+width]

        return BandSample(self.name, dst_raster, self.crs, dst_transform, self.nodata)

    def reproject(self, dst_crs, interpolation='nearest'):
        """ Change coordinate system (projection) of the band.
        It returns a new BandSample and does not alter the current object

        It is based on `rasterio.warp.reproject
        <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_,
        see for more variants of interpolation.

        Args:
            dst_crs: new CRS, may be in any form acceptable by rasterio, for example as EPSG code, string, CRS object; if dst_crs == `utm`, the appropriate UTM zone is used according to the center of the image
            interpolation: interpolation type as in rasterio,  `nearest`, `bilinear`, `cubic`, `lanzsos` or others

        Returns:
            BandSample: a new instance with changed CRS.
        """
        if isinstance(dst_crs, str) and dst_crs == 'utm':
            dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        else:
            dst_crs = dst_crs if isinstance(dst_crs, CRS) else CRS.from_user_input(dst_crs)

        # Old rasterio compatibility: a separate check for validity
        if not dst_crs.is_valid:
            raise rasterio.errors.CRSError('Invalid CRS {} given'.format(dst_crs))

        dst_transform, dst_width, dst_height = calculate_default_transform(
            self.crs, dst_crs, self.width, self.height, *self.bounds)

        new_raster = np.empty(shape=(1, dst_height, dst_width), dtype=self.dtype)

        reproject(
            self._raster, new_raster,
            src_transform=self.transform,
            dst_transform=dst_transform,
            src_crs=self.crs,
            dst_crs=dst_crs,
            resampling=getattr(Resampling, interpolation))

        return BandSample(self.name, new_raster, dst_crs, dst_transform, self.nodata)

    def reproject_to_utm(self, interpolation='nearest'):
        """
        Alias of :obj:`BandSample.reproject` method with automatic Band utm zone determining
        """
        dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        return self.reproject(dst_crs, interpolation=interpolation)

    def resample(self, dst_res=None, dst_shape=None, interpolation='nearest'):
        """ Change spatial resolution of the sample, resizing the raster according to the new resolution.
        dst_res should be specified, otherwise the destination transform will be equal to the source.
        If dst_shape is not specified, it is calculated from dst_res,
        but it can be specified to override it and get the desired output shape

        It is based on `rasterio.warp.reproject
        <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_,
        see for more variants of interpolation.

        Args:
            dst_res (Tuple[float, float]): new resoluton, georeferenced pixel size for the new band
            dst_shape: new shape of the resampled raster, can override calculated new shape
            interpolation: interpolation type as in rasterio,  `nearest`, `bilinear`, `cubic`, `lanzsos` or others

        Returns:
            a new resampled BandSample.
        """
        transform = self.transform if dst_res is None else Affine(dst_res[1],
                                                                  self.transform.b,
                                                                  self.transform.c,
                                                                  self.transform.d,
                                                                  - dst_res[0],
                                                                  self.transform.f)

        if dst_res is not None and dst_shape is None:
            target_height = int(self.height * self.res[0] / dst_res[0])
            target_width = int(self.width * self.res[1] / dst_res[1])
        elif dst_shape is not None:
            target_height = dst_shape[1]
            target_width = dst_shape[2]
        else:
            target_height = self.height
            target_width = self.width

        new_raster = np.empty(shape=(1, target_height, target_width), dtype=self.dtype)

        reproject(
            self._raster, new_raster,
            src_transform=self.transform,
            dst_transform=transform,
            src_crs=self.crs,
            dst_crs=self.crs,
            resampling=getattr(Resampling, interpolation))

        return BandSample(self._name, new_raster, self.crs, transform, self.nodata)

    def numpy(self):
        """
        A numpy representation of the raster, without metadata

        Returns: Sample's raster data as a numpy array
        """
        return self._raster

    def generate_samples(self, width, height):
        """
        A generator for sequential sampling of the whole sample, similar to Band,
        used for the windowed processing of the raster data.

        Args:
            width (int): dimension of sample in pixels and step along `X` axis
            height (int): dimension of sample in pixels and step along `Y` axis

        Yields:
            BandSample: sequential samples of the specified dimensions
        """
        for x in range(0, self.width, width):
            for y in range(0, self.height, height):
                yield self.sample(y, x, height, width)



import os
import rasterio

from ..dataset import BandCollection


def split(src_fp, dst_fp, channels, exist_ok=True):
    """Split multi-band tiff to separate bands

    This is necessary to prepare the source multi-band data for use with the BandCollection

    Args:
        src_fp: file path to multi-band tiff
        dst_fp: destination path to band collections
        channels: names for bands
        exist_ok:

    Returns:
        BandCollection

    """

    # create directory for new band collection
    os.makedirs(dst_fp, exist_ok=exist_ok)

    # parse extension of bands
    ext = src_fp.split('.')[-1]

    # open existing GeoTiff
    with rasterio.open(src_fp) as src:

        assert len(channels) == src.count

        profile = src.profile
        profile.update({
            'count': 1,
        })

        # write each band separately
        dst_pathes = []
        for n in range(src.count):

            dst_band_path = os.path.join(dst_fp, channels[n] + '.{}'.format(ext))
            with rasterio.open(dst_band_path, 'w', **profile) as dst:
                dst.write(src.read(n+1), 1)

            dst_pathes.append(dst_band_path)

    return BandCollection(dst_pathes)


import numpy as np

EPS = 10e-12


# ============================= Object-wise metrics ==================================

def iou(gt_feature, pr_feature):
    intersection = gt_feature.intersection(pr_feature).area
    union = gt_feature.union(pr_feature).area
    return (intersection + 10e-12) / (union + 10e-12)


def collection_iou(gt_fc, pr_fc) -> np.array:
    
    scores = []
    
    for feature in gt_fc:

        proposed_features = pr_fc.intersection(feature)
        feature_scores = [iou(feature, f) for f in proposed_features]
        if feature_scores:
            max_iou_score = max([iou(feature, f) for f in proposed_features])
        else:
            max_iou_score = 0
        scores.append(max_iou_score)
    
    return np.array(scores)
        

def confusion_matrix(gt_fc, pr_fc, iou_threshold=0.5):
    
    scores = collection_iou(gt_fc, pr_fc)
    tp = sum(scores > iou_threshold)
    fp = len(pr_fc) - tp
    fn = len(gt_fc) - tp
    
    return tp, fp, fn


def mAPxx(gt_fc, pr_fc, thresholds, beta=1) -> float:
    
    scores = collection_iou(gt_fc, pr_fc)
    mAPs = []
    
    for threshold in thresholds:
        tp = sum(scores > threshold)
        fp = len(pr_fc) - tp
        fn = len(gt_fc) - tp
        
        mAP_ = ((1 + beta ** 2) * tp + EPS) / ((1 + beta ** 2) * tp + beta ** 2 * fn + fp + EPS)
        mAPs.append(mAP_)
        
    return np.mean(mAPs)


def mAP(gt_fc, pr_fc, iou_threshold=0.5, beta=1) -> float:
    return mAPxx(gt_fc, pr_fc, thresholds=[iou_threshold], beta=beta)


def mAP50(gt_fc, pr_fc, beta=1) -> float:
    return mAPxx(gt_fc, pr_fc, beta=beta, thresholds=[0.5])


def mAP75(gt_fc, pr_fc, beta=1) -> float:
    return mAPxx(gt_fc, pr_fc, beta=beta, thresholds=[0.75])

def mAP5095(gt_fc, pr_fc, beta=1) -> float:
    thresholds = np.linspace(0.5, 0.95, 10)
    return mAPxx(gt_fc, pr_fc, beta=beta, thresholds=thresholds)


# ============================= Area metrics ==================================

def _union(fc):
    a = fc[0]
    for f in fc:
        a = a.union(f)
    return a


def area_iou(gt_fc, pr_fc):
    gt = _union(gt_fc)
    pr = _union(pr_fc)

    intersection = gt.intersection(pr).area
    union = gt.union(pr).area

    return (intersection + EPS) / (union + EPS)



import numpy as np


def add_mask(image, mask, channel='r', intensity=100):

        r, g, b = image.transpose(2,0,1)
        mask = mask.squeeze() * intensity

        if channel == 'r':
            r = r.astype(np.uint16) + mask.astype(np.uint16)
            r = np.clip(r, 0, 255).astype(np.uint8)

        if channel == 'g':
            g = g.astype(np.uint16) + mask.astype(np.uint16)
            g = np.clip(g, 0, 255).astype(np.uint8)

        if channel == 'b':
            b = b.astype(np.uint16) + mask.astype(np.uint16)
            b = np.clip(b, 0, 255).astype(np.uint8)

        image = np.stack((r, g, b)).transpose(1,2,0)

        return image


from .dataset import RandomDataset

from .raster import Band
from .raster import BandCollection
from .raster import BandSample
from .raster import BandCollectionSample

from .vector import Feature
from .vector import FeatureCollection

from .io import Predictor

from .transforms import polygonize
from .transforms import rasterize

from .utils import parse_directory


from functools import wraps

import keras.backend as K
from keras.losses import binary_crossentropy
from keras.losses import categorical_crossentropy
from keras.utils.generic_utils import get_custom_objects

from .metrics import iou_score, custom_iou_score
from .metrics import f1_score, custom_f_score


# ============================== Jaccard Losses ==============================

def jaccard_loss(gt, pr):
    return 1 - iou_score(gt, pr)


def _bce_jaccard_loss(gt, pr, bce_weight=1, smooth=1):

    iou_score = custom_iou_score(smooth=smooth)

    bce = K.mean(binary_crossentropy(gt, pr))
    jaccard_loss = 1 - iou_score(gt, pr)

    return bce_weight * bce + jaccard_loss


def _cce_jaccard_loss(gt, pr, bce_weight=1, class_weights=1, smooth=1):

    iou_score = custom_iou_score(class_weights=class_weights, smooth=smooth)

    cce = categorical_crossentropy(gt, pr) * class_weights
    cce = K.mean(cce)
    jaccard_loss = 1 - iou_score(gt, pr)
    return bce_weight * cce + jaccard_loss


def bce_jaccard_loss(gt, pr):
    return _bce_jaccard_loss(gt, pr, bce_weight=1, smooth=1)


def cce_jaccard_loss(gt, pr):
    return _cce_jaccard_loss(gt, pr, bce_weight=1, class_weights=1, smooth=1)


### Custom jaccard losses

def custom_jaccard_loss(class_weights=1, smooth=1):
    metric = custom_iou_score(class_weights=class_weights, smooth=smooth)
    def loss(gt, pr):
        return 1 - metric(gt, pr)
    return loss

def custom_bce_jaccard_loss(bce_weight=1, smooth=1):
    def loss(gt, pr):
        return _bce_jaccard_loss(gt, pr, bce_weight=bce_weight, smooth=smooth)
    return loss


def custom_cce_jaccard_loss(bce_weight=1, class_weights=1, smooth=1):
    def loss(gt, pr):
        return _cce_jaccard_loss(gt, pr, bce_weight=bce_weight, class_weights=class_weights, smooth=smooth)
    return loss


# Update custom objects
get_custom_objects().update({
    'jaccard_loss': jaccard_loss,
    'bce_jaccard_loss': bce_jaccard_loss,
    'cce_jaccard_loss': cce_jaccard_loss,
})





import os
import rasterio
from tqdm import tqdm
from multiprocessing.pool import ThreadPool
from threading import Lock

from .raster import Band
from .raster import BandCollection


class SequentialSampler:

    def __init__(self, band_collection, channels, sample_size, bound=0):
        """ Iterate over BandCollection sequentially with specified shape (+ bounds)
        Args:
            band_collection: BandCollection instance
            channels: list of str, required channels with required order
            sample_size: (height, width), size of `pure` sample in pixels (bounds not included)
            bound: int, bounds in pixels added to sample
        Return:
            Iterable object (yield SampleCollection instances)
        """

        self.band_collection = band_collection
        self.sample_size = sample_size
        self.bound = bound
        self.channels = channels
        self.blocks = self._compute_blocks()

    def __len__(self):
        return len(self.blocks)

    def __getitem__(self, i):
        block = self.blocks[i]
        sample = (self.band_collection
                  .ordered(*self.channels)
                  .sample(block['y'], block['x'], block['height'], block['width']))
        return sample, block

    def _compute_blocks(self):

        h, w = self.sample_size
        blocks = []
        height = h + 2 * self.bound
        width = w + 2 * self.bound

        for y in range(- self.bound, self.band_collection.height, h):
            for x in range(- self.bound, self.band_collection.width, w):

                rigth_x_bound = max(self.bound,
                                    x + width - self.band_collection.width)
                bottom_y_bound = max(self.bound,
                                    y + height - self.band_collection.height)

                blocks.append({'x': x,
                               'y': y,
                               'height': height,
                               'width': width,
                               'bounds':
                                   [[self.bound, bottom_y_bound], [self.bound, rigth_x_bound]],
                               })
        return blocks


class SampleWindowWriter:

    def __init__(self, fp, shape, transform, crs, nodata, dtype='uint8'):
        """ Create empty `Band` (rasterio open file) and write blocks sequentially

        Args:
            fp: file path of created Band
            shape: (height, width), size of band in pixels
            transform: rasterio Affine object
            crs: rasterio CRS or epsg core of coordinate system
            nodata: value of pixels without data
            dtype: str, one of rasterio data types

        Returns:
            when closed return `Band`

        Examples:
            ```python
            # create band
            bc = BandCollection(['/path/to/RED.tif', '/path/to/GRN.tif'])

            src = SequentialSampler(bc, channels, (1024, 1024), 512)
            dst = SampleWindowWriter('./test.tif', src.shape, **bc.profile)

            for sample, block in src:
                # read raster
                raster = sample.ordered('RED').numpy()

                # transform raster
                raster += 1

                # write raster
                dst.write(raster, **block)

            # close file when all data precessed
            created_band = dst.close()
            ```
        """
        self.fp = fp
        self.shape = shape
        self.transform = transform
        self.nodata = nodata
        self.crs = crs
        self.dtype = dtype
        self.dst = self.open()

    @property
    def height(self):
        return self.shape[0]

    @property
    def width(self):
        return self.shape[1]

    def open(self):
        return rasterio.open(self.fp, 'w', driver='GTiff', transform=self.transform, crs=self.crs,
                                 height=self.height, width=self.width, count=1,
                                 dtype=self.dtype, nodata=self.nodata)

    def close(self):
        self.dst.close()
        return Band(self.fp)

    def write(self, raster, y, x, height, width, bounds=None):
        """ Writes the specified raster into a window in dst
        The raster boundaries can be cut by 'bounds' pixels to prevent boundary effects on the algorithm output.
        If width and height are not equal to size of raster (after the bounds are cut), which is not typical,
        the raster is stretched to the window size (width and height)

        Args:
            raster: numpy array to be written into dst
            x: begin position of window
            y: begin position of window
            width: size of window
            height: size of window
            bounds: [[,][,]] - number of pixels to cut off from each side of the raster before writing

        Returns:

        """


        if bounds:
            raster = raster[bounds[0][0]:raster.shape[0]-bounds[0][1], bounds[1][0]:raster.shape[1]-bounds[1][1]]
            x += bounds[1][0]
            y += bounds[0][0]
            width = width - bounds[1][1] - bounds[1][0]
            height = height - bounds[0][1] - bounds[0][0]
        self.dst.write(raster, 1, window=((y, y+height), (x, x+width)))


class SampleCollectionWindowWriter:

    def __init__(self, directory, channels, shape, transform, crs, nodata, dtype='uint8'):
        """ Create empty `Band` (rasterio open file) and write blocks sequentially

        Args:
            direcory: directory path of created BandCollection
            channels: channel names of created BandCollection
            shape: (height, width), size of band in pixels
            transform: rasterio Affine object
            crs: rasterio CRS or epsg core of coordinate system
            nodata: value of pixels without data
            dtype: str, one of rasterio data types

        Returns:
            when closed return `BandCollection`

        Examples:
            ```python
            # create band
            bc = BandCollection(['/path/to/RED.tif', '/path/to/GRN.tif'])

            src = SequentialSampler(bc, channels, (1024, 1024), 512)
            dst = SampleCollectionWindowWriter('./test.tif', src.shape, **bc.profile)

            for sample, block in src:
                # read raster
                raster = sample.numpy()

                # transform raster
                raster += 1

                # write raster
                dst.write(raster, **block)

            # close file when all data precessed
            created_bc = dst.close()
            ```
        """
        if directory:
            os.makedirs(directory, exist_ok=True)

        self.fps = [os.path.join(directory, channel + '.tif') for channel in channels]
        self.channels = channels
        self.shape = shape
        self.transform = transform
        self.nodata = nodata
        self.crs = crs
        self.dtype = dtype
        self.writers = self.open()

    def open(self):
        writers = []
        for fp in self.fps:
            writers.append(
                SampleWindowWriter(fp, self.shape, self.transform,
                                   self.crs, self.nodata, self.dtype)
            )
        return writers

    def write(self, raster, y, x, height, width, bounds=None):
        for i in range(len(self.channels)):
            self.writers[i].write(raster[i], y, x, height, width, bounds=bounds)

    def close(self):
        bands = [w.close() for w in self.writers]
        return BandCollection(bands)


class Predictor:

    def __init__(self, input_channels, output_labels, processing_fn,
                 sample_size=(1024, 1024), bound=256, n_workers=1, verbose=True, **kwargs):
        """

        Args:
            input_channels: list of str, names of bands/channels
            output_labels: list of str, names of output classes
            processing_fn: callable, function that take as an input `SampleCollection`
                and return raster with shape (output_labels, H, W)
            sample_size: (height, width), size of `pure` sample in pixels (bounds not included)
            bound: int, bounds in pixels added to sample
            
        Returns:
            processed BandCollection
        """

        self.input_channels = input_channels
        self.output_labels = output_labels
        self.processing_fn = processing_fn
        self.sample_size = sample_size
        self.bound = bound
        self.kwargs = kwargs
        self.n_workers = n_workers
        self.verbose = verbose
        self.lock = Lock()

    def _threaded_processing(self, args):
        self._processing(*args)

    def _processing(self, sample, block, dst):
        raster = self.processing_fn(sample)
        with self.lock:
            dst.write(raster, **block)

    def process(self, bc, output_directory):

        src = SequentialSampler(bc, self.input_channels, self.sample_size, self.bound)
        dst = SampleCollectionWindowWriter(output_directory, self.output_labels,
                                           bc.shape[1:], **bc.profile, **self.kwargs)

        args = ((sample, block, dst) for sample, block in src)
        blocks_num = ((bc.shape[1] + self.bound) // self.sample_size[0] + 1) *\
                     ((bc.shape[2] + self.bound) // self.sample_size[1] + 1)

        if self.n_workers > 1:
            with ThreadPool(self.n_workers) as p:
                with tqdm(total=blocks_num, disable=(not self.verbose)) as pbar:
                    for _ in p.imap(self._threaded_processing, args):
                        pbar.update()
        else:
            with tqdm(args, total=blocks_num, disable=(not self.verbose)) as data:
                for sample, block, dst in data:
                    self._processing(sample, block, dst)

        return dst.close()



import os
from abc import ABC

import numpy as np
from .band import Band

from .geoobject import GeoObject
from ..coords import get_utm_zone


class BandCollection(GeoObject):
    """
    A collection of :obj:`Band` s with the same crs, transform and shape.

    It is intended to contain a multi-band image
    and associated raster masks. All the bands must have the same resolution, so if the initial images have
    different resolution, they must be resampled previously.
    Every band must be stored in a separate file; if the initial data is in one file,
    use :meth:`aeronet.converters.split.split` to split it to separate files

    Args:
        bands: list of `Band` or list of file paths
    """

    def __init__(self, bands):

        super().__init__()

        self._bands = [band if isinstance(band, Band) else Band(band) for band in bands]
        if not self.is_valid:
            del self
            raise ValueError('Bands are not suitable for collection. '
                             'CRS, transform or shape are different!')

    def __repr__(self):
        names = [b.name for b in self._bands]
        return '<BandCollection: {}>'.format(names)

    def __getitem__(self, item):
        return self._bands[item]

    def __len__(self):
        return self.count

    # ======================== PROPERTY BLOCK ========================

    @property
    def crs(self):
        return self._bands[0].crs

    @property
    def transform(self):
        return self._bands[0].transform

    @property
    def nodata(self):
        return self._bands[0].nodata

    @property
    def height(self):
        return self._bands[0].height

    @property
    def width(self):
        return self._bands[0].width

    @property
    def count(self):
        return len(self._bands)

    @property
    def bounds(self):
        return self._bands[0].bounds

    @property
    def shape(self):
        return self.count, self._bands[0].height, self._bands[0].width

    @property
    def res(self):
        return self._bands[0].res

    @property
    def is_valid(self):
        """Check if all bands have the same resolution, shape and coordinate system"""
        if len(self._bands) < 2:
            res = True
        else:
            first = self._bands[0]
            rest = self._bands[1:]
            res = all(first.same(other) for other in rest)

        return res

    # ======================== PRIVATE METHODS ========================

    def _get_band(self, name):
        for b in self._bands:
            if b.name == name:
                return b
            elif len(b.name) > len(name):  # legacy datasets support
                if b.name.endswith('_{name}'.format(name=name)):
                    return b
            # in all other cases raise error
        raise NameError('No sample with name {name}.'.format(name=name))

    # ======================== PUBLIC METHODS  ========================

    def append(self, other):
        """Add a band to collection, checking it to be compatible by shape, transform and crs"""
        if all(other.same(band) for band in self._bands):
            self._bands.append(other)
        else:
            raise ValueError('Band is not suitable for collection. '
                             'CRS, transform or shape are different!')
        return

    def sample(self, y, x, height, width):
        """
        Sample memory object from BandCollection.

        Args:
            x: int, top left corner `X` offset in pixels
            y: int, top left corner `Y` offset in pixels
            width: int, width of samples in pixels
            height: int, height of samples in pixels

        Returns:
            a new :obj:`BandCollectionSample` containing the specified spatial subset of the BandCollection
        """
        samples = [band.sample(y, x, height, width) for band in self._bands]
        return BandCollectionSample(samples)

    def ordered(self, *names):
        """
        Creates a new object, containing the specified bands in the specific order.

        Args:
            *names: order of names

        Returns:
            reordered `BandCollection`
        """
        ordered_bands = [self._get_band(name) for name in names]
        return BandCollection(ordered_bands)

    def reproject(self, dst_crs, directory=None, interpolation='nearest'):
        """
        Reprojects every Band of the collection, see :meth:`Band.reproject` and returns a new reprojected BandCollection
        """
        if directory:
            os.makedirs(directory, exist_ok=True)
        r_bands = []
        for band in self:
            fp = os.path.join(directory, band.name + '.tif') if directory else None
            r_band = band.reproject(dst_crs, fp=fp, interpolation=interpolation)
            r_bands.append(r_band)
        return BandCollection(r_bands)

    def reproject_to_utm(self, directory=None, interpolation='nearest'):
        """
        Alias of `reproject` method with automatic utm zone determining
        """
        dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        return self.reproject(dst_crs, directory=directory, interpolation=interpolation)

    def resample(self, dst_res, directory=None, interpolation='nearest'):
        """
        Resamples every Band of the collection, see :meth:`Band.resample` and returns a new reprojected BandCollection
        """

        if directory:
            os.makedirs(directory, exist_ok=True)
        r_bands = []
        for band in self:
            fp = os.path.join(directory, band.name + '.tif') if directory else None
            r_band = band.resample(dst_res, fp=fp, interpolation=interpolation)
            r_bands.append(r_band)
        return BandCollection(r_bands)

    def generate_samples(self, height, width):
        """
        A generator for sequential sampling of the whole BandCollection, used for the windowed reading of the raster.
        It allows to handle and process large files without reading them at once in the memory.

        Args:
            width (int): dimension of sample in pixels and step along `X` axis
            height (int): dimension of sample in pixels and step along `Y` axis

        Yields:
            BandCollectionSample: sequential samples of the specified dimensions
        """

        for x in range(0, self.width, width):
            for y in range(0, self.height, height):
                yield self.sample(y, x, height, width)

    def numpy(self):
        return self.sample(0, 0, self.height, self.width).numpy()


class BandCollectionSample(GeoObject):
    """
    A collection of :obj:`BandSample` , which are also
    """

    def __init__(self, samples):
        super().__init__()
        # todo: add is_valid check
        self._samples = samples

    def __repr__(self):
        names = [b.name for b in self._samples]
        return '<BandCollectionSample: {}>'.format(names)

    def __getitem__(self, item):
        return self._samples[item]

    def __len__(self):
        return self.count

    # ======================== PROPERTY BLOCK ========================

    @property
    def crs(self):
        return self._samples[0].crs

    @property
    def transform(self):
        return self._samples[0].transform

    @property
    def res(self):
        return self._samples[0].res

    @property
    def width(self):
        return self._samples[0].width

    @property
    def height(self):
        return self._samples[0].height

    @property
    def count(self):
        """ Number of bands (layers) in the collection"""
        return len(self._samples)

    @property
    def shape(self):
        return self.count, self.height, self.width

    @property
    def nodata(self):
        return self._samples[0].nodata

    @property
    def bounds(self):
        return self._samples[0].bounds

    @property
    def is_valid(self):
        """
        Check if all bands have the same resolution, shape and coordinate system
        """
        if len(self._samples) < 2:
            res = True
        else:
            first = self._samples[0]
            rest = self._samples[1:]
            res = all(first.same(other) for other in rest)

        return res

    # ======================== PRIVATE METHODS ========================

    def _get_sample(self, name):
        for s in self._samples:
            if s.name == name:
                return s
            elif len(s.name) > len(name):  # legacy datasets support
                if s.name.endswith('_{name}'.format(name=name)):
                    return s
            # in all other cases raise error
        raise NameError('No sample with name {name}.'.format(name=name))

    # ======================== PUBLIC METHODS  ========================

    def append(self, obj):
        """
        Add sample to collection, checking it to be compatible by shape, transform and crs
        """
        if all(obj.same(band) for band in self._samples):
            self._samples.append(obj)
        else:
            raise ValueError('Band is not suitable for collection. '
                             'CRS, transform or shape are different!')
        return

    def sample(self, y, x, height, width):
        samples = [obj.sample(y, x, height, width) for obj in self._samples]
        return BandCollectionSample(samples)

    def reproject(self, dst_crs, interpolation='nearest'):
        """
        Reprojects every BandSample of the collection, see :meth:`BandSample.reproject`
        and returns a new reprojected BandCollectionSample
        """
        reprojected_samples = [s.reproject(dst_crs, interpolation) for s in self._samples]
        return BandCollectionSample(reprojected_samples)

    def reproject_to_utm(self, interpolation='nearest'):
        """
        Alias of `reproject` method with automatic utm zone determining
        """
        dst_crs = get_utm_zone(self.crs, self.transform, (self.height, self.width))
        return self.reproject(dst_crs, interpolation=interpolation)

    def resample(self, dst_res=None, dst_shape=None, interpolation='nearest'):
        """
        Reprojects every BandSample of the collection, see :meth:`BandSample.reproject`
        and returns a new reprojected BandCollectionSample
        """
        resamples = [s.resample(dst_res, dst_shape, interpolation) for s in self._samples]
        return BandCollectionSample(resamples)

    def numpy(self):
        return np.stack([x.numpy() for x in self._samples], 0)

    def ordered(self, *names):
        """
        Creates a new object, containing the specified bands in the specific order.

        Args:
            *names: order of names

        Returns:
            reordered `BandCollectionSample`
        """
        ordered_bands = [self._get_sample(name) for name in names]
        return BandCollectionSample(ordered_bands)

    def save(self, directory, extension='.tif', **kwargs):
        """
        Saves every BandSample in the specified deirectory, see :meth:`BandSample.save`
        """
        os.makedirs(directory, exist_ok=True)
        for s in self._samples:
            s.save(directory, extension, **kwargs)

    def generate_samples(self, height, width):
        """
        A generator for sequential sampling of the BandCollectionSample similar to BandCollection,
        used for the windowed processing of the raster data.

        Args:
            width (int): dimension of sample in pixels and step along `X` axis
            height (int): dimension of sample in pixels and step along `Y` axis

        Yields:
            BandCollectionSample: sequential samples of the specified dimensions
        """
        for x in range(0, self.width, width):
            for y in range(0, self.height, height):
                yield self.sample(y, x, height, width)
